/*
** iplog_scan.c - iplog scan/flood detector.
** Copyright (C) 1999 Ryan McCabe <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
*/

#include <config.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <pthread.h>
#include <rwlock.h>

#include <iplog.h>
#include <iplog_scan.h>
#include <iplog_options.h>

#define ENDSCAN \
	"%s mode expired for %s - received a total of %lu packets (%lu bytes)."

extern u_long flags;

__inline__ static bool no_res(void) {
	return (0);
}

static struct scan_list scan_table[SCAN_TSIZE];

static const struct scan_info scan[] = {
	{ "TCP: port scan",			tcp_res,	PS_TIMEOUT,		PS_THOLD	},
	{ "TCP: null scan",			tcp_res,	NS_TIMEOUT,		NS_THOLD	},
	{ "TCP: FIN scan",			tcp_res,	FS_TIMEOUT,		FS_THOLD	},
	{ "TCP: Xmas scan",			tcp_res,	XS_TIMEOUT,		XS_THOLD	},
	{ "UDP: scan/flood",		udp_res,	UDP_TIMEOUT,	UDP_THOLD	},
	{ "ICMP/UDP: smurf attack",	no_res,		SMURF_TIMEOUT,	SMURF_THOLD	},
	{ "ICMP: ping flood",		icmp_res,	PING_TIMEOUT,	PING_THOLD	}
};

bool check_scan(u_char type, addr_t saddr, u_long len) {
	u_long hash;
	addr_t addr;
	struct scan_data *cur;
	struct scan_t *data;
	bool ret = false, found = false, log_scan = false;

	if (type == SCAN_SMURF)
		addr = saddr & 0xffffff;
	else
		addr = saddr;

	hash = HASH(addr, SCAN_TSIZE);

	rwl_readlock(&scan_table[hash].lock);
	for (cur = scan_table[hash].head ; cur != NULL ; cur = cur->next) {
		if (cur->addr == addr) {
			found = true;
			data = &cur->type[type];

			pthread_mutex_lock(&data->lock);

			data->bytes += len;
			data->expire = time(NULL) + scan[type].timeout;

			if (++data->count >= scan[type].threshold) {
				ret = true;
				if (data->logged == false) {
					data->logged = true;
					log_scan = true;
				}
			}

			pthread_mutex_unlock(&data->lock);
			break;
		}
	}
	rwl_readunlock(&scan_table[hash].lock);

	if (found == false) {
		rwl_writelock(&scan_table[hash].lock);
		if (scan_table[hash].cnt < SCAN_MAXENT) {
			struct scan_data temp, *t_cur;

			memset(&temp, 0, sizeof(temp));
			temp.addr = addr;
			t_cur = list_prepend(&temp, (void **) &scan_table[hash].head,
						sizeof(temp));
			scan_table[hash].cnt++;
			data = &t_cur->type[type];

			pthread_mutex_lock(&data->lock);

			data->count = 1;
			data->bytes = len;
			data->expire = time(NULL) + scan[type].timeout;

			pthread_mutex_unlock(&data->lock);
		} /* FIXME: else [log], remove oldest, add new entry */
		rwl_writeunlock(&scan_table[hash].lock);
	}

	if (log_scan == true) {
		u_char buf[256];
		struct in_addr in;

		in.s_addr = addr;

		mysyslog("%s detected from %s",	scan[type].name,
			host_lookup(&in, scan[type].resolv(), buf, sizeof(buf)));

		return (true);
	}

	return (ret);
}

void expire_scans(void) {
	size_t i, j;
	struct scan_data *cur = NULL;
	bool remove;
	u_char buf[256];
	struct in_addr in;

	for (i = 0 ; i < SCAN_TSIZE ; i++) {
		rwl_writelock(&scan_table[i].lock);
		cur = scan_table[i].head;

		for (remove = true ; cur != NULL ;) {
			for (j = 0 ; j < SCAN_TOTAL ; j++) {
				pthread_mutex_lock(&cur->type[j].lock);

				if (cur->type[j].count != 0 &&
					time(NULL) >= cur->type[j].expire)
				{
					if (cur->type[j].count >= scan[j].threshold) {
						in.s_addr = cur->addr;

						mysyslog(ENDSCAN, scan[j].name,
							host_lookup(&in, scan[j].resolv(), buf,sizeof(buf)),
							cur->type[j].count, cur->type[j].bytes);
					}

					cur->type[j].expire = 0;
					cur->type[j].logged = 0;
					cur->type[j].count = 0;
					cur->type[j].bytes = 0;
				} else if (cur->type[j].count != 0)
					remove = false;

				pthread_mutex_unlock(&cur->type[j].lock);
			}

			if (remove == true) {
				cur = list_delete(cur, (void **) &scan_table[i].head);
				scan_table[i].cnt--;
			} else
				cur = cur->next;
		}

		rwl_writeunlock(&scan_table[i].lock);
	}
}

#ifdef HAVE_PTHREAD_CANCEL
void destroy_scan_table(void) {
	size_t i;

	for (i = 0 ; i < SCAN_TSIZE ; i++) {
		if (scan_table[i].head != NULL)
			list_destroy(scan_table[i].head, NULL);
		memset(&scan_table[i], 0, sizeof(struct scan_list));
	}
}
#endif

/* vim:ts=4:sw=8:tw=0 */
